// project.e
//
// Thu, 07/23/1992 12:55:11
//
// Purpose:
//
//		This extension is designed as an aid to the individual who
//		has code or information within a project directory tree or
//		forest which is possibly widely scattered over the
//		development environment.
//
//		The primary purpose of this extension is to create an
//		environment in which all file operations (such as opens,
//		greps, etc. see details below) will work over all files in a
//		project.  The method by which this is accomplished is by the
//		implementation of a set of project_get_file routines which
//		are similar to the built in Epsilon get_file routines.  This
//		requires the creation of a completion mechanism for file
//		completion.  Using these get_file routines allows the
//		replacement or hook of those commands dealing with files.
//		The current values of the new project variables are stored
//		within the session files saved during exit and restored on
//		startup.  When files are specified on the command line the
//		previous session is restored first and then the new files are
//		added.
//
//		This extension defines two new variables which are set by the
//		user to specify the root directory location of the current
//		project and all directories which are a part of that project.
//		The project directory and its associated parts may all be
//		absolute or relative path names (which allows multiple
//		branches of a project on one machine at different absolute
//		locations but at similar relative locations for auxiliary
//		files).  I call the project a forest rather than simply a
//		tree because absolute files are allowed in the multiple
//		directory variable, project_directories, to allow the use of
//		one or more global locations at which files may be located
//		(ie \c700\include).  See the description of project_directory
//		and project_directories below to get more detail in setting
//		these variables.
//
//		Four additional variables are defined, three to modify the
//		behavior of session save and restore (session_preserve_local,
//		session_search_local and session_always_restore) and one to
//		keep the current active session file for use by this and
//		other extensions (session_current_file).  See the descriptions
//		below for details.
//
//		The commands defined within this extension are replacements
//		for commands which need to be used across an entire project
//		forest of files but currently are limited to one directory
//		location per invocation.  These commands are implemented to
//		perform exactly as the Epsilon equivalents if the project
//		variables are set to NULL and thus should present no problems
//		when used outside of any particular project environment.
//		Even if the project variables are set, these commands will
//		always match files within the current subdirectory.  This
//		ensures that even those who have only one session state file
//		can use these commands transparently.
//
//		The project support is spread about in several commands:
//
//			project_find_file() provides a replacement for
//			find_file() which has full completion across the project
//			forest.
//
//			project_pluck_file() provides a simple extension to
//			project_find_file() which uses the (possible) file name
//			currently under the cursor as a default file name so that
//			traversal into a referenced file (ie #include "foo.h") is
//			made quite a bit simpler.
//
//			project_grep() is a project replacement for Epsilon's
//			grep() command.  This grep will search all files across
//			the entire defined project which match the file specs
//			given by the user.
//
//			project_tag_files() is a project replacement for
//			Epsilon's tag_files() command.  This command enables you
//			to tag all files of interest (ie. *.c) within your entire
//			project.
//
//			Similarly the following are replacements for commands
//			which allow completion over a project tree.
//				project_copy_to_file()
//				project_insert_file()
//				project_load_bytes()
//				project_load_file()
//				project_visit_file()
//				project_write_file()
//				project_write_region()
//
// Usage:
//
//		This extension is very simple to configure.  Add project.doc
//		to the end of edoc, load project.b with the load-bytes
//		command, bind the commands you are going to use to the non
//		project specific version of the commands and save the state
//		using the command write-state.
//
//		The global project variables used by this extension
//		(project_directory and project_directories) are saved within
//		the session state file when exiting Epsilon and restored when
//		entering Epsilon.  This is accomplished via the session hooks
//		defined within session.e in the Epsilon distribution and do
//		not require that they be saved in the state file.
//
// Details:
//
//		The behavior variable ignore_file_extensions is not currently
//		dealt with.  The completion mechanism for the project forest
//		will show all matches *regardless* of extension.
//
//		A project forest always contains the current subdirectory.
//		This forces these routines to behave as their non project
//		counterparts when the project_director* variables are empty
//		strings.
//
// Some future directions:
//
//		1) Deal correctly with the ignore_file_extensions variable.
//
// The modifications to this code are dedicated to the public domain
// with the caveat that Lugaru is welcome to use this within their
// distribution source code which is supplied with Epsilon.  Use it
// any way you want for whatever purposes.  Portions are Copyright
// Lugaru Software...
//
//
//	John Kercheval
//	15563 10th Ave NE
//	Seattle, WA 98155-6209
//	INTERNET: johnk@wrq.com
//	Compu$erve: 72450,3702
//
//
// Modification History:
//
//		Wed, 07/08/1992 23:44:04 V1.0 initial release.
//
//		Thu, 07/09/1992 22:43:46 V1.1 add session variables and
//		command line expansion support.  Adds significant usability
//		to the session mechanism and behaves much better for project
//		oriented and standard behavior edits.  Place the default
//		bindings for the replacement commands.
//
//		Thu, 07/09/1992 23:49:01 V1.2 add the commands
//		project_write_file, project_copy_to_file,
//		project_insert_file, project_write_region, load_file,
//		load_bytes and select_tag_file to cover the whole gambit of
//		routines using the get_file() routine.
//
//		Tue, 07/14/1992 21:59:13 V1.3 add read only check to session
//		restore hook so that files which have change read only status
//		will be correctly brought into the system.
//
//		Wed, 07/22/1992 21:02:26 V1.4 add more documentation to
//		header and add verbosity to the write_session command.  The
//		command find_project_file subroutine now works as does the
//		find_file routine in Epsilon with the except that completion
//		on dot directories behaves as does other directory
//		completions rather than giving an empty completion list.
//		Those not using the project features of this extension should
//		see behavior nearly identical with Epsilon's default.  Place
//		C-X C-F as project_find_file for compatibility reasons.
//		pluck_project_find_file is still a bindable command.  Find
//		file commands will open all files which fit the pattern and
//		show the last file found in the current buffer.  Fix bug in
//		absolute path check, break absolute path check into a
//		subroutine.
//
//		Thu, 07/23/1992 12:38:48 V1.5 write session will now save the
//		current session to the startup session file without
//		prompting.  If a numeric argument is given before the
//		write_session command then a file name prompt is given with
//		the default in the current project directory.
//
//		Mon, 07/27/1992 20:47:39 V1.6 fix bug opening new files using
//		project_find_file and project_pluck_file.  Use a unique
//		delimiter to fencepost the session information for the
//		project.  This will avoid collision with other session save
//		and restore hooks.  This is a dirty trick because it blows
//		away your previous project information, but this new approach
//		is relatively immune to competing session hooks and is order
//		independent.
//
//		Wed, 07/29/1992 21:34:36 V1.7 change session variables so
//		they all begin with session, this makes them easy to find via
//		completion.  Add new variable session_local_search to support
//		new directory hierarchy search for a project session file.
//		Fix a subtle bug in write_session and save_session which if
//		interrupted would store an invalid session_current_file
//		creating a problem on exit and storing the session in the
//		Epsilon default rather than in the startup project session.
//		Fix bug in read only buffer parsing init.
//
//		Tue, 09/08/1992 20:28:14 V1.8 remove the readonly file check
//		to reflect the fix performed by Lugaru for V6.02.
//
//		Thu, 09/24/1992 13:38:35 V1.9 write_session would save the
//		session always to the Epsilon default session file when it
//		should have been saving to the current active session file.
//
//		Sat, 02/06/1993 10:47:26 V2.0 remove session routines
//		(except for the session hooks), they are no longer
//		needed.  Update for V6.5 of Epsilon.
//
//
// some amount of Epsilon original code so...
//
/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*   "EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.          *
*                                                                       *
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.  *
*                                                                       *
* Limited permission is hereby granted to reproduce and modify this     *
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in  *
* any such reproduction or modification.                                *
************************************************************************/


#include <eel.h>

#ifndef BOOL
#define BOOL int
#define TRUE 1
#define FALSE 0
#endif

#define MAX_BUF		256

//
// A list of subdirectories or absolute directories that project-grep
// will search in.  It is semicolon separated list just like PATH
// environment variable.  If a component is not absolute, it will be
// interpreted as relative to project_directory variable.  For example,
// suppose that project_directory is "c:\project" and this variable
// is set to "foo;d:\bar", then the following directories will be
// searched, in order (note the implicit first search of the current
// subdirectory),
//
//  the current directory
//  c:\project\foo
//  d:\bar
//
user char project_directories[MAX_BUF] = "";

//
// This is the root pathname for the project grep command.
// Directories mentioned in project_directories are interpreted as
// relative to this.  It may be absolute or relative, for example,
// "c:\project" or "project".  This variable should be relative if
// you have multiple branches for versions of a project on one system
// or you have multiple session files laying around.
//
user char project_directory[FNAMELEN] = "";

//
// This variable is TRUE (1) if the system should always restore
// files and variables found in the session file even if the command
// line contains some specific file names on it.  Note that if the
// command line does specify files then the window configuration is
// not preserved, just the files and associated variables, marks,
// bookmarks, etc.  If no files are specified on the command line or
// if this variable is FALSE (0) then session restore works as
// the default.
//
user BOOL session_always_restore = TRUE;


////////////////////////////////////////////////////////////////////////////
//
// is_absolute will return true if the passed path is absolute.
//

BOOL is_absolute(path)
	char *path;
{
	if (strlen(path) &&
		(path[0] == '\\' ||
		 path[0] == '/' ||
		 path[1] == ':'))
		return TRUE;
	else
		return FALSE;
}


////////////////////////////////////////////////////////////////////////////
//
// maybe_prepend_path will combine the prefix and path if path is not
// absolute and return in ret, do not exceed length characters.
//

maybe_prepend_path(prefix, path, ret, length)
	char *prefix;				// path element to prepend
	char *path;					// relative path element
	char *ret;					// return string
	int length;					// maximum length of ret
{
	if (!strlen(path)) {
		strncpy(ret, prefix, length - 1);
		return;
	}
	if (is_absolute(path)) {
		strncpy(ret, path, length - 1);
		return;
	}
	strncpy(ret, prefix, length - 1);
	if (path[strlen(path) - 1] == '\\' ||
		path[strlen(path) - 1] == '/' )
		strncat(ret, path, length - strlen(prefix) - 1);
	else {
		strncat(ret, "\\", length - strlen(prefix) - 1);
		strncat(ret, path, length - strlen(prefix));
	}
}


////////////////////////////////////////////////////////////////////////////
//
// maybe_append_path will combine the postfix and path if path is not
// a file name (ie. contains a '.') and return in ret, do not exceed
// length characters.
//

maybe_append_path(postfix, path, ret, length)
	char *postfix;				// path element to append
	char *path;					// relative path element
	char *ret;					// return string
	int length;					// maximum length of ret
{
	if (!strlen(path)) {
		strncpy(ret, postfix, length - 1);
		return;
	}
	strncpy(ret, path, length - 1);
	if (path[strlen(path) - 1] == '\\' ||
		path[strlen(path) - 1] == '/' )
		strncat(ret, postfix, length - strlen(path) - 1);
	else {
		strncat(ret, "\\", length - strlen(path) - 1);
		strncat(ret, postfix, length - strlen(path));
	}
}


////////////////////////////////////////////////////////////////////////////
//
// make_session_hook and restore_session_hook save the current
// project variables to the session file when it is written or read
// and then call any other defined session hooks.
//

#define PROJECT_START_DELIM	"<<SES!@!PROJECT_START>>"
#define PROJECT_END_DELIM	"<<SES!@!PROJECT_END>>"

new_project_make_session_hook(b)
{
	save_var bufnum = b;
	
	point = size();
	
	buf_printf(b, "\n%s", PROJECT_START_DELIM);
	buf_printf(b, "\n%s\n%s",
			   project_directory,
			   project_directories);
	buf_printf(b, "\n%s\n", PROJECT_END_DELIM);

	project_make_session_hook(b);
}

REPLACE_FUNC("project", "make_session_hook")

new_project_restore_session_hook(b)
{
	int old_point = point;

	save_var bufnum = b;
	
	//
	// obtain the project variables
	//
	point = 0;
	if (!search(1, PROJECT_START_DELIM)) {
		point = old_point;
	} else {
		nl_forward();			 // move to first variable
		grab_line(b, project_directory);
		grab_line(b, project_directories);
		nl_forward();			 // move beyond final delimiter
	}

	//
	// call the previous hook
	//
	project_restore_session_hook(b);
}

REPLACE_FUNC("project", "restore_session_hook")


////////////////////////////////////////////////////////////////////////////
//
// get_project_directory will return all project subdirectories one
// at a time and return FALSE when no more subdirectories are found.
// Flags is set to MATCHSTART when the first subdirectory is required
// (reset).  The return string will not exceed length characters.
//

//
// This is an index used during the parsing of the project path
//
int project_directory_index;

BOOL
get_project_directory(ret, length, flags)
	char *ret;					// return string
	int length;					// maximum length of return string
	int flags;					// standard completion flags
{
	char *begin_directory;		// the start of this element
	char *end_directory;		// the end of this element
	char ret_root[FNAMELEN];	// the intermediate return value
	//
	// Are we just starting the match?  If so, return the current
	// directory which is always a part of the current project.
	//
	if (flags & STARTMATCH) {
		project_directory_index = 0;
		*ret = '\0';
		return TRUE;
	}
	//
	// If we are at the end of the list return FALSE
	//
	if (project_directory_index >= strlen(project_directories) - 1) {
		*ret = '\0';
		return FALSE;
	}
	//
	// Move over leading semicolons and place at beginning of
	// possible match
	//
	begin_directory = project_directories + project_directory_index;
	while (*begin_directory == ';') {
		begin_directory++;
	}
	//
	// If at the end of the string we are done
	//
	if (*begin_directory == '\0') {
		project_directory_index = begin_directory - project_directories;
		*ret = '\0';
		return FALSE;
	}
	//
	// We found something, copy it over and adjust statics accordingly
	//
	if ((end_directory = strstr(begin_directory, ";")) == NULL) {
		strncpy(ret_root, begin_directory, length - 1);
		project_directory_index = strlen(project_directories);
	} else {
		if (end_directory - begin_directory < length) {
			strncpy(ret_root, begin_directory, end_directory - begin_directory);
		} else {
			strncpy(ret_root, begin_directory, length - 1);
		}
		project_directory_index = end_directory - project_directories;
	}
	maybe_prepend_path(project_directory, ret_root, ret, length);
	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
//
// project_file_match will return all project files which match
// partial one at a time and return NULL when no more files are
// found.  Flags is set to FM_PREFIX, FM_EXACT when the first match
// is required (reset).  The return value will include the project
// directory information.
//

//
// This is an intermediate path variable used to track the completion
// state for the project file lister and match routine.
//
char project_current_directory[FNAMELEN] = "";

//
// This is the initial startup flags used.  This is stored for reuse
// as a parameter to file_match on successive project directories.
//
int start_flags = 0;

char *
project_file_match(partial, flags)
	char *partial;				// return string
	int flags;					// standard completion flags
{
	BOOL found = FALSE;			// sentinal
	BOOL start = FALSE;			// file_match start state
	BOOL project_expand = TRUE;	// true if partial is not absolute

	char project_current_file[FNAMELEN];	// full completion string
	char *ret;					// return value

	//
	// Init
	//
	if (flags & FM_EXACT || flags & FM_PREFIX) {
		get_project_directory(project_current_directory, FNAMELEN, STARTMATCH);
		start_flags = flags;
		start = TRUE;
	}
	while (!found) {
		//
		// If we are expanding "\\" or "\" don't bother with the
		// project expansions.  If this is the first time through
		// then return a pointer to partial.
		//
		if (!strcmp("\\", partial) ||
			!strcmp("/", partial)) {
			strcpy(project_current_file, partial);
 			if (start)
				return project_current_file;
			else
				return NULL;
		}
		//
		// If the string ends with one of the dot directories there
		// is only one official match.
		//
		if ((strlen(partial) > 1 &&
			partial[strlen(partial)-1] == '.' &&
			(partial[strlen(partial)-2] == '.' ||
			 partial[strlen(partial)-2] == '\\' ||
			 partial[strlen(partial)-2] == '/')) ||
			(strlen(partial) == 1 && partial[0] == '.')) {
			strcpy(project_current_file, partial);
			strcat(project_current_file, "\\");
 			if (start)
				return project_current_file;
			else
				return NULL;
		}
		//
		// If we are expanding an absolute or primary relative path
		// then don't bother with the project expansions.
		//
		if (is_absolute(partial) ||
			!strncmp(".", partial, 1) ||
			!strncmp("..", partial, 2)) {
			strcpy(project_current_file, partial);
			project_expand = FALSE;
		} else
			maybe_append_path(partial, project_current_directory,
							  project_current_file, FNAMELEN);

		//
		// If a match is forthcoming then leave the loop, otherwise
		// enumerate the project directory and try again.
		//
		if ((ret = file_match(project_current_file,
							  start ? start_flags : 0)) != NULL)
			found = TRUE;
		else if (!project_expand)
			return NULL;
		else if (!get_project_directory(project_current_directory, FNAMELEN, 0))
			return NULL;
		else
			start = TRUE;
	}
	return ret;
}


////////////////////////////////////////////////////////////////////////////
//
// project_file_full_finder will return all project files which match
// partial one at a time and return FALSE when no more files are
// found.  Flags is set to MATCHSTART when the first match is
// required (reset).  The return value will include the project
// directory information.
//

char *
project_file_full_finder(partial, flags)
	char *partial;				// return string
	int flags;					// standard completion flags
{
	int file_flags = 0;

	if (flags & STARTMATCH)
		if (flags & EXACTONLY)
			file_flags = FM_EXACT;
		else
			file_flags = FM_PREFIX;
	return project_file_match(partial, file_flags);
}


////////////////////////////////////////////////////////////////////////////
//
// project_file_finder will return all project files which match
// partial one at a time and return FALSE when no more files are
// found.  Flags is set to MATCHSTART when the first match is
// required (reset).  The return value will not include project
// directory information.
//

char *
project_file_finder(partial, flags)
	char *partial;				// return string
	int flags;					// standard completion flags
{
	char *ret;

	//
	// get the next match
	//
	ret = project_file_full_finder(partial, flags);

	//
	// Remove the leading project directory information to allow the
	// completion mechanism to complete without a listing.
	//
	if (ret != NULL) {
		if (is_absolute(partial))
			return ret;
		else {
			ret += strlen(project_current_directory);
			if (*ret == '\\' ||
				*ret == '/')
				ret++;
		}
	}
	return ret;
}


////////////////////////////////////////////////////////////////////////////
//
// project_file_lister will place all project files matching prefix
// in the buffer mbuf using the project finder function.  The
// parameter finder is ignored in this routine.
//

project_file_lister(prefix, finder, mbuf)
	char *prefix;				// the completion string
	char *(*finder) ();			// a suggested finder function

{
	char *current_match;		// temporary char pointer
	save_var bufnum = zap("-list-temp");	// previous active buffer

	finder = finder;			// get rid of pesky warning

	if ((current_match = project_file_full_finder(prefix, STARTMATCH)) != NULL) {
		stuff(current_match);
		stuff("\n");
		while ((current_match = project_file_full_finder(prefix, 0)) != NULL) {
			stuff(current_match);
			stuff("\n");
		}
	}
	buf_zap(mbuf);
	do_buffer_sort(bufnum_to_name(mbuf), 0, 0);
	bufnum = mbuf;
	delete_buffer("-list-temp");
	point = 0;
	delete_final_newlines();
}


////////////////////////////////////////////////////////////////////////////
//
// get_project_file will use file completion to obtain a file name
// somewhere in the project directory structure.  This routine will
// not necessarily return an absolute path name as get_file() and
// get_file_dir() do.
//

get_project_file(response, prompt, def)
	char *response;				// return value
	char *prompt;				// the initial prompt
	char *def;					// the default string
{
	char fullprompt[FNAMELEN];	// built up prompt name
	int len = strlen(prompt);	// length of final prompt

	save_var completion_lister = project_file_lister;

	strcpy(fullprompt, prompt);
	if (*def && strlen(def) * 2 < screen_cols - len - 6)
		len += sprintf(fullprompt + len, " [%s]", def);
	strcpy(fullprompt + len, ": ");
	comp_read(response, fullprompt, project_file_finder,
			  CAUTIOUS | FOLDFILE, def);
	if (!*response)
		strcpy(response, def);
}


////////////////////////////////////////////////////////////////////////////
//
// get_project_file_dir will use file completion to obtain a file
// pattern somewhere in the project directory structure.  This
// routine will not necessarily return an absolute path name as
// get_file() and get_file_dir() do.  This will use the current
// buffers directory as the default.
//

get_project_file_dir(response, prompt)
	char *response;				// return value
	char *prompt;				// the initial prompt
{
	char def[FNAMELEN], *p;		// the default response

	save_var completion_lister = project_file_lister;

	if (is_dired_buf())		/* get directory name from dired */
		dired_getdir(def);
	else {
		strcpy(def, filename);
		absolute(def);
		if (*filename)	/* want only directory part */
			*get_tail(def, 1) = 0;
		p = def + strlen(def);	/* ptr to end of string */
		if (!is_path_separator(p[-1])) {
			*p++ = path_sep; /* make sure it ends with slash */
			*p = 0;
		}
	}

	comp_read(response, prompt, project_file_finder,
			  CAUTIOUS | FOLDFILE, def);
}


////////////////////////////////////////////////////////////////////////////
//
// project_find_file will use get_project_file to obtain a file name
// and it will open the obtained file.
//

command
project_find_file()
	on  cx_tab[CTRL('f')]
{
	char fname[FNAMELEN];		// file pattern to read in
	char openfname[FNAMELEN];	// The file to open
	char *full_fname;			// project match return

	iter = 0;
	tiled_only();

	get_project_file_dir(fname, "Find file: ");

	//
	// get the full path of the completed file, if the file is a new
	// one then just open it using the fname return.
	//
	if (*fname) {
		full_fname = project_file_match(fname, FM_EXACT);
		if (!full_fname) {
			strcpy(openfname, fname);
			absolute(openfname);
			do_find(openfname, has_arg ? !strip_returns.default
					: strip_returns.default);
		} else {
			while (full_fname != NULL) {
				strcpy(openfname, full_fname);
				absolute(openfname);
				do_find(openfname, has_arg ? !strip_returns.default
						: strip_returns.default);
				full_fname = project_file_match(fname, 0);
			}
		}
	} else
		dired_one(fname);
}


////////////////////////////////////////////////////////////////////////////
//
// pluck_file will obtain the file name at the current buffers cursor
// location, read it in and display it as the default for a file find
// file routine.
//

#define FILE_PATTERN "[a-zA-Z0-9_\x80-\x9A\xA0-\xA5%.%(%){}%\\~!@#%$%%%^&%-%+=]+"

command
project_pluck_file()
{
	char fname[FNAMELEN];		// The file pattern to open
	char openfname[FNAMELEN];	// The file to open
	char *full_fname;			// The completed project file
	char def[FNAMELEN];			// The plucked default file

	save_var point,
	mark;

	iter = 0;
	point--;

	//
	// Pluck the file name from the current buffer and complete with
	// that file name as the default.  No 8.3 error checking...
	//
	re_search(1, FILE_PATTERN);
	re_search(-1, FILE_PATTERN);
	grab(point, matchstart, def);
	restore_vars();
	get_project_file(fname, "Find file", def);

	//
	// get the full path of the completed file, if the file is a new
	// one then just open it using the fname return.
	//
	full_fname = project_file_match(fname, FM_EXACT);
	if (!full_fname) {
		strcpy(openfname, fname);
		absolute(openfname);
		do_find(openfname, has_arg ? !strip_returns.default
				: strip_returns.default);
	} else {
		while (full_fname != NULL) {
			strcpy(openfname, full_fname);
			absolute(openfname);
			do_find(openfname, has_arg ? !strip_returns.default
					: strip_returns.default);
			full_fname = project_file_match(fname, 0);
		}
	}
}


////////////////////////////////////////////////////////////////////////////
//
// project_copy_to_file will copy the current buffer to a file
// without changing the current buffer filename or modification
// history.  Modified from copy_to_file() in files.e in Epsilon
// distribution.
//

command
project_copy_to_file()
	on  reg_tab[FCTRL(7)]
{
	int mod = modified;
	char tmp[FNAMELEN],
	    orig[FNAMELEN];

	iter = 0;
	strcpy(orig, filename);
	get_project_file(tmp, "Copy to file", filename);
	filename = tmp;
	do_save_file(1, 0, 0);
	filename = orig;
	modified = mod;
}


////////////////////////////////////////////////////////////////////////////
//
// project_visit_file will replace the current buffer with the
// contents of the asked for file name.  Modified from visit_file()
// in files.e in Epsilon distribution.
//

command
project_visit_file()
	on  cx_tab[CTRL('V')]
{
	char tmp[FNAMELEN];
	char resp[80];

	iter = 0;
	get_project_file(tmp, "Visit file", filename);
	absolute(tmp);
	if (!*tmp)
		return;
	if (!has_arg && modified) {	/* buffer need to be saved? */
		get_string(resp, "Save buffer? [y]");
		if (toupper(*resp) != 'N')
			do_save_file(1, 1, 1);
	}
	read_file(tmp, strip_returns);
}


////////////////////////////////////////////////////////////////////////////
//
// project_insert_file will ask for a file name and insert it at the
// current point in the current buffer.  Modified from insert_file()
// in files.e in Epsilon distribution.
//

command
project_insert_file()
	on  cx_tab['i']
{
	char file[FNAMELEN];

	iter = 0;
	get_project_file(file, "Insert file", _region_file);
	absolute(file);
	if (!*file)
		return;
	strcpy(_region_file, file);
	mark = point;		/* put mark before it */
	build_first = 1;
	do_insert_file(file, strip_returns);
}


////////////////////////////////////////////////////////////////////////////
//
// project_write_region will ask for a file name and save the current
// region to that file if one is given.  Modified from write_region()
// in files.e in Epsilon distribution.
//

command
project_write_region()
	on  cx_tab['w']
{
	char file[FNAMELEN];

	get_project_file(file, "Write region to file", _region_file);
	absolute(file);
	if (!*file)
		return;
		strcpy(_region_file, file);
	if (!write_part(file, strip_returns, point, mark))
			note("%s written.", file);
}


////////////////////////////////////////////////////////////////////////////
//
// project_write_file will ask for a file name and save buffer to
// that file if one is given.  Modified from write_file() in files.e
// in Epsilon distribution.
//

command
project_write_file()
	on  cx_tab[CTRL('W')]
{
	char tmp[FNAMELEN];

	iter = 0;
	get_project_dir(tmp, "Write file: ");
	absolute(tmp);
	if (*tmp) {
		filename = tmp;
		return do_save_file(1, 0, 1);
	}
	return 0;
}


////////////////////////////////////////////////////////////////////////////
//
// project_load_file will load definitions from a file.  Modified
// from load_file() in control.e in Epsilon distribution.
//

command
project_load_file()
{
	char fname[80];

	get_project_file(fname, "Load definitions from file", "");
	absolute(fname);
	if (*fname) {
		if (!find_index("do-load-file"))
			load_from_path("store");
		do_load_file(fname);
	}
}


////////////////////////////////////////////////////////////////////////////
//
// project_load_bytes will load byte codes from a file.  Modified
// from load_bytes() in load.e in Epsilon distribution.
//

char _prev_byte_name[80];

command
project_load_bytes()
	on  reg_tab[FKEY(3)]
{
	char fname[80],
	    rel[80];

	get_project_file(fname, "Load bytecodes from file", _prev_byte_name);
	if (!*fname)
		return;

	absolute(fname);
	strcpy(get_extension(fname), byte_extension);
	strcpy(_prev_byte_name, fname);
	load_commands(fname);
	relative(fname, rel);
	say("%s loaded.", rel);
	iter = 0;
}


////////////////////////////////////////////////////////////////////////////
//
// project_do_grep performs the actual bulk of the work for the
// project_grep command.  This routine is a modified piece from
// grep.e in the Epsilon distribution.
//

project_do_grep(buf, pat, flags, filepat)
	char *buf,
	   *pat,
	   *filepat;
{
	int res = 0,
	    err = 0,
	    cnt = 0,
	    start,
	    tempbuf = 0,
	    orig,
	    before_head;
	char *s,
	    rel[FNAMELEN],
	    abs[FNAMELEN],
	    search_type[80];

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (size())
		bprintf("\n");			/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	bprintf("Found 0 lines with %s\"%s\" in %s\n",
			search_type, pat, filepat);
	start = point;
	display_column = 0;			/* horizontally scroll this window */
	grep_mode();
	if (!(s = project_file_match(filepat, FM_EXACT | FM_NO_DIRS)))
		error("No matches");
	save_var readonly_warning = 0;
	save_var regex_first_end = 1;	/* settings for fastest searches */
	save_var regex_shortest = 1;

	orig = bufnum;
	for (; s; s = project_file_match(filepat, 0)) {
		strcpy(abs, s);
		if (*abs)
			absolute(abs);
		relative(abs, rel);
		noteput("%d matches, searching %s...", cnt, rel);
		if (grep_keeps_files)
			find_it(abs, strip_returns.default);
		else if (!look_file(s)) {
			if (!tempbuf)
				tempbuf = tmp_buf();
			bufnum = tempbuf;
			if (err = err_file_read(abs, strip_returns.default))
				break;
		}
		res = copy_matches(name_to_bufnum(buf), flags, pat, rel);
		window_bufnum = orig;
		if (res < 0)			/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			point = start;
			delete(point, before_head);
			bprintf("Found %d lines with %s\"%s\" in %s\n",
					cnt, search_type, pat, filepat);
			start = point;
			point = size();
			maybe_refresh();
		}
	}
	window_bufnum = orig;
	point = start;
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (err)
		quick_abort();
	if (res == -DSBAD)
		error("illegal search pattern");
	else if (res != -DSABORT && cnt) {
		locate_window(buf, "");
		modified = 0;
		point = start;
	}
}


////////////////////////////////////////////////////////////////////////////
//
// project_grep will perform a grep on all completion files entered
// by the user in the current project.  Modified from grep() in
// grep.e in Epsilon distribution.
//

command
project_grep()
	on  reg_tab[FALT(7)]
{
	int flags;
	char filepat[FNAMELEN];

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);			/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ", UNKNOWN | REGEX | GREP);
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	get_project_file(filepat, "In file(s)",
					 *_grep_files ? _grep_files : filename);
	if (!*filepat)
		return;
	strcpy(_grep_files, filepat);
	project_do_grep("grep", (flags & REGEX) ? _default_regex_search
					: _default_search, flags, filepat);
}


////////////////////////////////////////////////////////////////////////////
//
// project_tag_files will perform a tag operation on all completion
// files entered by the user in the current project.  Modified from
// tag_files() in tags.e in Epsilon distribution.
//

command
project_tag_files()
	on  cx_tab[ALT('.')]
{
	char pat[FNAMELEN],
	    abs[FNAMELEN],
		*s;

	init_tags();
	get_project_file(pat, "Add/update tags for files matching", filename);
	iter = 0;
	if (!(s = project_file_match(pat, FM_EXACT | FM_NO_DIRS)))
		error("No matches");
	for (; s; s = project_file_match(pat, 0)) {
		strcpy(abs, s);
		if (*abs)
			absolute(abs);
		delete_tags(abs);
		tag_a_file(abs);
	}
	do_save_tags();
	say("%s tagged.", pat);
}


////////////////////////////////////////////////////////////////////////////
//
// project_select_tag_file will switch to a particular tags file.
// Modified from select_tag_file() in tags.e in Epsilon distribution.
//

command
project_select_tag_file()
	on  cx_tab[ALT(',')]
{								/* switch to a particular tags file */
	char file[FNAMELEN],
	   *def = initial_tag_file;
	int old = bufnum;

	do_save_tags();
	if (exist("-tags")) {
		bufname = "-tags";
		def = filename;
		bufnum = old;
	}
	get_project_file(file, "Tag file", def);
	absolute(file);
	load_tags(file);
	say("Tags loaded from %s", file);
}


////////////////////////////////////////////////////////////////////////////
//
// load_cmd_line_files will restore the epsilon session before
// reading in the command line files and assigning them to windows.
// This subroutine will open all files in the project tree matching
// the command line patterns.
//

load_cmd_line_files(i)
{								/* start at argv[i] */
	int wins = 0,
	    tot_size = window_height,
	    line;
	char fname[FNAMELEN],
	    abs[FNAMELEN],
		*s;

	//
	// restore session if wanted but remove window configuration set
	// as a result
	//
	if (session_always_restore)
		load_session();

	iter = 0;
	one_window();

	//
	// read in the file from the project specification.
	//
	line = -1;
	for (; i < argc; i++) {
		if (argv[i][0] == '+')
			line = strtoi(argv[i] + 1, 10);
		else {
			strcpy(fname, argv[i]);
			s = expand_wildcards ? project_file_match(fname, FM_EXACT)
				: fname;
			if (!s)				/* new file */
				s = fname;
			do {
				strcpy(abs, s);
				if (*abs)
					absolute(abs);
				if (wins < 3 && wins++ > 0) {
					split_window();
					window_height = 1000;
				}
				do_find(abs, strip_returns.default);
				if (line > 0)
					go_line(line);
			} while (expand_wildcards
					 && (s = project_file_match(fname, 0)));
			line = -1;
		}
	}
	for (; window_number > 0; window_number--)
		window_height = tot_size / number_of_windows();
}
